import tensorflow as tf
import tensorflow_probability as tfp
import keras.backend as K
tfd = tfp.distributions


class Losses:
    def __init__(self, loss_fn, observation_noise, beta=0, z_mean=None, z_log_sigma=None):
        self.loss = self._get_loss(loss_fn)
        self.beta = beta
        self.observation_noise = observation_noise
        self.z_mean = z_mean
        self.z_log_sigma = z_log_sigma
        if self.beta > 0:
            assert z_mean != None, "z_mean should be defined!"
            assert z_log_sigma != None, "z_log_sigma should be defined"

    def _get_loss(self, loss_fn):
        losses = {"binary": self._binary,
                  "normal": self._normal,
                  "normalDiag": self._normalDiag,
                  "bernoulli": self._bernoulli}
        return losses[loss_fn]

    """
      Making it custom metric to be able to feed it to Keras API - actually no need for y_true, y_pred
    """

    def KL_divergence(self, y_true, y_pred):
        if self.beta > 0:
            return - self.beta * 0.5 * K.mean(
                1 + self.z_log_sigma - K.square(self.z_mean) - K.exp(self.z_log_sigma), axis=-1)
        return 0
    """
      For binarized input with KL term (!)
    """

    def _binary(self, x_true, x_reco):
        return -tf.nn.sigmoid_cross_entropy_with_logits(
            labels=x_true, logits=x_reco) + self.KL_divergence(None, None)

    def _bernoulli(self, x_true, x_reco):
        return -tf.reduce_mean(tfd.Bernoulli(x_reco)._log_prob(x_true)
                               ) + self.KL_divergence(None, None)

    """
      For non binarized input with KL term(!)
    """

    def _normal(self, x_true, x_reco):
        return -tf.reduce_mean(
            tfd.Normal(x_reco, scale=self.observation_noise)._log_prob(x_true)
        ) + self.KL_divergence(None, None)

    def _normalDiag(self, x_true, x_reco):
        return -tf.reduce_mean(
            tfd.MultivariateNormalDiag(
                x_reco, scale_identity_multiplier=self.observation_noise)._log_prob(x_true)) + self.KL_divergence(None, None)



from .ladder_vae import LadderVAE
from keras.layers import Input, Conv2D, ReLU, Dense, MaxPooling2D, Flatten, Reshape, Deconv2D
from keras.models import Model
import numpy as np


class ConvLadderVAE(LadderVAE):
    def encoder1(self):
        input_img = Input(shape=self.input_shape[1:])
        x = Conv2D(256, (2, 2), padding='same')(input_img)
        x = ReLU()(x)
        x = MaxPooling2D((2, 2), padding='same')(x)
        x = Conv2D(128, (2, 2), padding='same')(x)
        x = ReLU()(x)
        x = MaxPooling2D((2, 2), padding='same')(x)
        x = Conv2D(128, (2, 2), padding='same')(x)
        x = ReLU()(x)
        x = Flatten()(x)
        encoded = Dense(self.latent_dim1)(x)
        encoder = Model(input_img, encoded)
        return encoder

    def encoder2(self):
        input_img = Input(shape=(self.latent_dim1,))
        x = Dense(256)(input_img)
        x = Reshape(target_shape=(16, 16, 1))(x)
        x = Conv2D(128, (2, 2), padding='same')(x)
        x = ReLU()(x)
        x = MaxPooling2D((2, 2), padding='same')(x)
        x = Conv2D(64, (2, 2), padding='same')(x)
        x = ReLU()(x)
        x = Flatten()(x)
        encoded = Dense(self.latent_dim2)(x)
        encoder = Model(input_img, encoded)
        return encoder

    def decoder2(self):
        latent2 = Input(shape=(self.latent_dim2,))
        x = Dense(256)(latent2)
        x = Reshape(target_shape=(16, 16, 1))(x)
        x = Conv2D(128, (2, 2), padding='same')(x)
        x = ReLU()(x)
        x = MaxPooling2D((2, 2), padding='same')(x)
        x = Conv2D(64, (2, 2), padding='same')(x)
        x = ReLU()(x)
        x = Flatten()(x)
        x = Dense(self._mean_variance_input_shape)(x)
        reco = ReLU()(x)
        decoder = Model(latent2, reco)
        return decoder

    def decoder1(self):
        latent1 = Input(shape=(self.latent_dim1,))
        x = Dense(1024)(latent1)
        x = Reshape(target_shape=(32, 32, 1))(x)
        x = Conv2D(256, (2, 2), padding='same')(x)
        x = ReLU()(x)
        x = MaxPooling2D((2, 2), padding='same')(x)
        x = Conv2D(128, (2, 2), padding='same')(x)
        x = ReLU()(x)
        x = MaxPooling2D((2, 2), padding='same')(x)
        x = Conv2D(128, (2, 2), padding='same')(x)
        x = ReLU()(x)
        x = Flatten()(x)
        x = Dense(np.prod(self.input_shape[1:]))(x)
        reco = Reshape(target_shape=self.input_shape[1:])(x)
        decoder = Model(latent1, reco)
        return decoder



from abc import abstractmethod
from keras.layers import Input, Dense, Lambda
from keras.models import Model
from keras.optimizers import Adam
from ..losses import Losses
from abc import abstractmethod
import tensorflow as tf
import tensorflow_probability as tfp
tfd = tfp.distributions


class Encoder:
    def __init__(self, input_shape, latent_dim):
        self.input_shape = input_shape
        self.latent_dim = latent_dim
        self.BATCH_SIZE = self.input_shape[0]

    @abstractmethod
    def _encoder(self):
        pass

    @abstractmethod
    def _decoder(self):
        pass

    @abstractmethod
    def get_compiled_model(self, *args):
        pass



from ..losses import Losses
from abc import abstractmethod
from keras.layers import Input, Dense, ReLU, Lambda, Add
from keras.models import Model
import keras.backend as K
from keras.optimizers import RMSprop, Adam
import tensorflow as tf
import tensorflow_probability as tfp
tfd = tfp.distributions


class LadderVAE:
    def __init__(self, input_shape, latent_dim1, latent_dim2):
        self.input_shape = input_shape
        self.latent_dim1 = latent_dim1
        self.latent_dim2 = latent_dim2
        self.BATCH_SIZE = self.input_shape[0]
        self._mean_variance_input_shape = 1024  # arbitrary

    @abstractmethod
    def encoder1(self):
        pass

    @abstractmethod
    def encoder2(self):
        pass

    @abstractmethod
    def decoder2(self):
        pass

    @abstractmethod
    def decoder1(self):
        pass

    def mean_variance_model(self):
        inp = Input(shape=(self._mean_variance_input_shape,))
        mean, var = Dense(self.latent_dim1)(inp), Dense(self.latent_dim1)(inp)
        model = Model(inp, [mean, var])
        return model

    def _get_sigma(self, args):
        sigma1, sigma2 = args
        return K.pow(K.pow(sigma1, -2) + K.pow(sigma2, -2), -1)

    def _get_mean(self, args):
        mean1, sigma1, mean2, sigma2 = args
        return (mean1 * K.pow(sigma1, -2) + mean2 * K.pow(sigma2, -2)) * self._get_sigma([sigma1, sigma2])

    def _get_sigma_gen(self, args):
        sigma1 = args
        return K.pow(sigma1, 2)

    def _get_mean_gen(self, args):
        mean1, sigma1 = args
        return mean1 * K.pow(sigma1, -2) * self._get_sigma_gen(sigma1)

    def _sample(self, args):
        z_mean, z_sigma = args
        dist = tfd.Normal(loc=z_mean, scale=z_sigma)
        return dist.sample()

    def _reparametrize(self, args):
        z_mean, z_log_sigma = args
        epsilon = K.random_normal(
            shape=(self.BATCH_SIZE, self.latent_dim2), mean=0, stddev=1)
        return z_mean + K.exp(z_log_sigma) * epsilon

    def get_compiled_model(self, *args):
        loss_fn, lr, decay, self.observation_noise, self.beta = args
        input_img = Input(batch_shape=self.input_shape)

        encoder1 = self.encoder1()
        encoder2 = self.encoder2()
        decoder1 = self.decoder1()
        decoder2 = self.decoder2()
        mean_var_model_for_top_down_calc = self.mean_variance_model()

        d1 = encoder1(input_img)
        d2 = encoder2(d1)

        # Reparametrization deepest layer
        self.z2_mean, self.z2_log_sigma = Dense(
            self.latent_dim2)(d2), Dense(self.latent_dim2)(d2)

        self.z2 = Lambda(self._reparametrize, name="latent")(
            [self.z2_mean, self.z2_log_sigma])

        losses = Losses(loss_fn, self.observation_noise,
                        self.beta, self.z2_mean, self.z2_log_sigma)

        # Top down and bottom up mean and variance calculation
        self.z1_intermediate = decoder2(self.z2)

        self.z1_mean_TD, self.z1_sigma_TD = mean_var_model_for_top_down_calc(
            self.z1_intermediate)

        self.z1_mean_BU, self.z1_sigma_BU = Dense(
            self.latent_dim1)(d1), Dense(self.latent_dim1)(d1)

        # Combine mean and sigma
        self.z1_sigma = Lambda(self._get_sigma)(
            [self.z1_sigma_BU, self.z1_sigma_TD])

        self.z1_mean = Lambda(self._get_mean)(
            [self.z1_mean_TD, self.z1_sigma_TD, self.z1_mean_BU, self.z1_sigma_BU])

        # Samlpe z1!
        self.z1 = Lambda(self._sample, name="sampling_z1")(
            [self.z1_mean, self.z1_sigma])

        reco = decoder1(self.z1)

        model = Model(input_img, reco)
        model.compile(optimizer=RMSprop(lr=lr, decay=decay),
                      loss=losses.loss, metrics=[losses.KL_divergence])

        # Generative model
        latent_input = Input(shape=(self.latent_dim2,))
        gen2 = decoder2(latent_input)

        # Using same TD mean var generator as before
        gen_mean, gen_sigma = mean_var_model_for_top_down_calc(gen2)

        # Combine mean and sigma for generative model
        gen_sigma = Lambda(self._get_sigma_gen)(
            [gen_sigma])

        gen_mean = Lambda(self._get_mean_gen)(
            [gen_mean, gen_sigma])

        # Samlping
        gen2 = Lambda(self._sample)(
            [gen_mean, gen_sigma])

        gen_reco = decoder1(gen2)
        generative_model = Model(latent_input, gen_reco)

        # Model for latent inference
        z2 = self.z2
        latent_model = Model(input_img, outputs=[reco, z2])

        self.latent_dim = self.latent_dim2

        return model, generative_model, latent_model



from keras.preprocessing.image import ImageDataGenerator
import numpy as np
from .data_loader import DataLoader


class DataGenerator:
    def __init__(self, image_shape, batch_size, file_path='/data/scramtex_700_28px.pkl', binarize=False):
        self.DATA_LOADER = DataLoader(file_path, binarize)
        self.IMAGE_SHAPE = image_shape
        self.BATCH_SIZE = batch_size
        self.DATA_GENERATOR = ImageDataGenerator()
        self.TRAIN, self.TEST = self.DATA_LOADER.train_test_split()

    def flow(self):
        return self.DATA_GENERATOR.flow(self.TRAIN, self.TRAIN, batch_size=self.BATCH_SIZE)

    def contrast_flow(self):
        def train_generator(_it):
            while True:
                batch_x, batch_y = next(_it)
                yield self._contrast(batch_x, batch_y)
        return train_generator(self.flow())

    def flattened_flow(self):
        image_dim = np.prod(self.IMAGE_SHAPE)

        def train_generator(_it):
            while True:
                batch_x, batch_y = next(_it)
                yield batch_x.reshape(self.BATCH_SIZE, image_dim), batch_y.reshape(self.BATCH_SIZE, image_dim)
        return train_generator(self.flow())

    def flattened_contrast_flow(self):
        image_dim = np.prod(self.IMAGE_SHAPE)

        def train_generator(_it):
            while True:
                batch_x, batch_y = next(_it)
                batch_x, batch_y = self._contrast(
                    batch_x, batch_y)
                yield batch_x.reshape(self.BATCH_SIZE, image_dim), batch_y.reshape(self.BATCH_SIZE, image_dim)

        return train_generator(self.flow())

    def _contrast(self, images_x, images_y):
        contrasted_images_x = np.zeros(shape=images_x.shape)
        contrasted_images_y = np.zeros(shape=images_y.shape)
        for ind in range(images_x.shape[0]):
            r_contrast = np.random.rand() * 2.
            contrasted_images_x[ind] = np.clip(
                r_contrast * (images_x[ind] - 0.5) + 0.5, 0, 1)
            contrasted_images_y[ind] = np.clip(
                r_contrast * (images_y[ind] - 0.5) + 0.5, 0, 1)
        return contrasted_images_x, contrasted_images_y

    def validation_data(self):
        return self.TEST, self.TEST

    def flattened_validation_data(self):
        image_dim = np.prod(self.IMAGE_SHAPE)
        reshaped_test = self.TEST.reshape(self.TEST.shape[0], image_dim)
        return reshaped_test, reshaped_test
